//модуль для клиента
'use strict';
// это клиент

// Зависимости
let app = require('express')();
let timestamp = require('time-stamp');
let ip = require('ip');
let nanoid = require('nanoid');
let axios = require('axios');
const bodyParser   = require('body-parser');
const jwt        = require('jsonwebtoken');

// Конфигурация
const wsName = require('../../package.json').name;
const wsVersion = require('../../package.json').version;
const wsClientConfig = require('../../wsClientConfig.json');
const wsPort = wsClientConfig.wsPort;
const wsType = jwt.decode(wsCoreKey, {complete: true}).payload.wsType;

let apigwPath = null;
let tokenForReq = null;

const instanceID = nanoid(10);


// Чтение body и cookie
app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: true}));
//app.use(cookieParser());


//let systemName = jwt.decode(wsCoreKey, {complete: true}).payload.systemName;

// Характеристики инстанса в сессии сокет-соединения
let socketToGlobalProvider = null;

let awaitedRequests = [];

socketToGlobalProvider.on('error', (error) => {
    msLog(`Получено событие об ошибке сокет-соединения`, 'err');
});


let globalApiGwPath = '';

function setGlobalApiGwPath(newPath){
    globalApiGwPath = newPath
}

function checkGlobalApiGwPath(){
    return globalApiGwPath !== '';
}

function clearGlobalApiGwPath(){
    globalApiGwPath = '';
}






// При успешном коннекте
socketToGlobalProvider.on('connect', function() {
    //sendStartData();
    //console.log('Подключились к socketServerToGlobalProvider');
    msLog(`Подключились к socketServerToGlobalProvider`, 'ok');
    //socket.emit('wsStart', 'wsStartData');
    // console.log('Подключились к socketServerToGlobalProvider');
    // console.log(socketToGlobalProvider.io.opts.query);
});

socketToGlobalProvider.on('setGateWayPath', function(data) {
    //wsPortForRequests
    //wsPortForEvents
    setGlobalApiGwPath(data.globalApiGwPath);
    msLog(`Получено событие для установки нового gateWayPath`, 'ok');
    //console.log(data)
});

socketToGlobalProvider.on('delGateWayPath', function() {
    msLog(`Получено событие об удалении gateWayPath`, 'ok');
    clearGlobalApiGwPath();
    //wsPortForRequests
    //wsPortForEvents
});

socketToGlobalProvider.on('preparedNewGateWayPath', function() {
    msLog(`Получено сообщение о необходимости смены gateWayPath`, 'ok');
    socketToGlobalProvider.emit('getGateWayPath');
});




// функция для общего логирования
function msLog(logString, mark) {
    const msMark = mark || 'ok' ;
    console.log(`${timestamp('[DD-MM-YYYY HH:mm:ss.ms]')}: (${msMark}) ${logString}` )
}

// Проверка что запрос пришел именно через apigw
app.use(function(req, res, next){
    if (req.method ==='OPTIONS'){
        next();
    }
    else {
        //console.log(`Промежуточная функция до вызова метода`);
        // if ((req.method ==='GET' ||  req.method ==='POST' || req.method ==='PUT') &&  ( req.headers.toinstanceid === instanceID)){
        if ((['GET', 'POST', 'PUT', 'DELETE'].includes(req.method)) &&  ( req.headers.toinstanceid === instanceID)){
            next();
        } else {
            msLog(`Обращение к сервису минуя apigw`, 'wrn');
            return res.send({result: 'error', reason: `Запрос к сервису ${wsName} минуя apigw`})
        }

    }
});

async function get (name, next) {
    awaitedRequests.push({methodName:name, methodType: 'GET'});
    next = (function() {
        let cached_function = next;
        return async function() {
            const reqId = arguments[0].query.s_reqId || arguments[0].query.u_reqId || 'none';
            msLog(`Начало обработки GET запроса ${name} с идентификатором ${reqId}`, 'info');
            await cached_function.apply(this, arguments); // use .apply() to call it
            //console.log(arguments[0].query.u_reqId);
            msLog(`Завершение обработки GET запроса ${name} с идентификатором ${reqId}`, 'info');
        };
    }());
    app.get(name, next)
}
async function post (name, next) {
    awaitedRequests.push({methodName:name, methodType: 'POST'});
    next = (function() {
        let cached_function = next;
        return async function() {
            const reqId = arguments[0].body.s_reqId || arguments[0].body.u_reqId || 'none';
            msLog(`Начало обработки POST запроса ${name} с идентификатором ${reqId}`, 'info');
            await cached_function.apply(this, arguments); // use .apply() to call it
            //console.log(arguments[0].query.u_reqId);
            msLog(`Завершение обработки POST запроса ${name} с идентификатором ${reqId}`, 'info');
        };
    }());
    app.post(name, next)
}
async function put (name, next) {
    awaitedRequests.push({methodName:name, methodType: 'PUT'});
    next = (function() {
        let cached_function = next;
        return async function() {
            const reqId = arguments[0].body.s_reqId || arguments[0].body.u_reqId || 'none';
            msLog(`Начало обработки PUT запроса ${name} с идентификатором ${reqId}`, 'info');
            await cached_function.apply(this, arguments); // use .apply() to call it
            //console.log(arguments[0].query.u_reqId);
            msLog(`Завершение обработки PUT запроса ${name} с идентификатором ${reqId}`, 'info');
        };
    }());
    app.put(name, next)
}
async function del (name, next) {
    awaitedRequests.push({methodName:name, methodType: 'DELETE'});
    next = (function() {
        let cached_function = next;
        return async function() {
            const reqId = arguments[0].query.s_reqId || arguments[0].query.u_reqId || 'none';
            msLog(`Начало обработки DELETE запроса ${name} с идентификатором ${reqId}`, 'info');
            await cached_function.apply(this, arguments); // use .apply() to call it
            //console.log(arguments[0].query.u_reqId);
            msLog(`Завершение обработки DELETE запроса ${name} с идентификатором ${reqId}`, 'info');
        };
    }());
    app.delete(name, next)
}

async function callGet (wsNameAndMethod, axiParams ) {
    const s_reqId = 's_'+ nanoid(6);
    if (!axiParams) {
        axiParams = {
            s_reqId:s_reqId
        }
    } else {
        axiParams.s_reqId = s_reqId;
    }
    msLog(`Осуществлен межсервисный GET запрос ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    let fullPath = `${apigwPath}/${wsNameAndMethod}`;
    const axiosRes = await axios.get(fullPath, {params:axiParams, headers: {tokenForReq:tokenForReq}});
    msLog(`Получен ответ от межсервисного GET запроса ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    return axiosRes.data
}

async function callPost (wsNameAndMethod, axiParams ) {
    const s_reqId = 's_'+ nanoid(6);
    if (!axiParams) {
        axiParams = {
            s_reqId:s_reqId
        }
    } else {
        axiParams.s_reqId = s_reqId;
    }
    msLog(`Осуществлен межсервисный POST запрос ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    let fullPath = `${apigwPath}/${wsNameAndMethod}`;
    const axiosRes = await axios.post(fullPath, axiParams, {headers: {tokenForReq:tokenForReq}});
    msLog(`Получен ответ от межсервисного POST запроса ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    return axiosRes.data
}

async function callPut (wsNameAndMethod, axiParams ) {
    const s_reqId = 's_'+ nanoid(6);
    if (!axiParams) {
        axiParams = {
            s_reqId:s_reqId
        }
    } else {
        axiParams.s_reqId = s_reqId;
    }
    msLog(`Осуществлен межсервисный PUT запрос ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    let fullPath = `${apigwPath}/${wsNameAndMethod}`;
    const axiosRes = await axios.put(fullPath, axiParams, {headers: {tokenForReq:tokenForReq}});
    msLog(`Получен ответ от межсервисного PUT запроса ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    return axiosRes.data
}

async function callDel (wsNameAndMethod, axiParams ) {
    const s_reqId = 's_'+ nanoid(6);
    if (!axiParams) {
        axiParams = {
            s_reqId:s_reqId
        }
    } else {
        axiParams.s_reqId = s_reqId;
    }
    msLog(`Осуществлен межсервисный DELETE запрос ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    let fullPath = `${apigwPath}/${wsNameAndMethod}`;
    const axiosRes = await axios.delete(fullPath, {params:axiParams, headers: {tokenForReq:tokenForReq}});
    msLog(`Получен ответ от межсервисного DELETE запроса ${wsNameAndMethod} с идентификатором ${s_reqId}`, 'info');
    return axiosRes.data
}

async function callAxiGet (axiPath, axiParams) {
    const s_reqId = 's_'+ nanoid(6);
    let allParams = {fromWsName:wsName, fromInstanceID:instanceID};
    if (axiParams !== undefined) {
        Object.assign(axiParams, allParams);
    }
    msLog(`Осуществлен межсистемный GET запрос к ${axiPath} с идентификатором ${s_reqId}`, 'info');
    const axiosRes = await axios.get(axiPath, {params:allParams});
    msLog(`Получен ответ от межсистемного GET запроса от ${axiPath} с идентификатором ${s_reqId}`, 'info');
    return axiosRes.data
}

async function callAxiPost (axiosPath, axiParams) {
    const s_reqId = 's_'+ nanoid(6);
    msLog(`Осуществлен межсистемный POST запрос к ${axiPath} с идентификатором ${s_reqId}`, 'info');
    const axiosRes = await axios.post(axiosPath, axiParams);
    msLog(`Получен ответ от межсистемного POST запроса от ${axiPath} с идентификатором ${s_reqId}`, 'info');
    return axiosRes.data
}



function start(params){
    let wsPort = params.wsPort;
    let wsCoreKey = params.wsCoreKey;

    let socketServer = jwt.decode(wsCoreKey, {complete: true}).payload.wsEventBusAddress;
    socketToGlobalProvider = require('socket.io-client')(socketServer,{
        query: {
            token: wsCoreKey, //jwt
            wsName: wsName,
            instanceID: instanceID,
            wsType: wsType,
        }
    });




    app.listen(wsPort, function(){
        msLog(`Сервис ${wsName} запущен на ${wsPort} порту`);
    });

}


module.exports = {
// Работа с запросами
    // Обработка входящих запросов
    get: get,
    post: post,
    put: put,
    del: del,
    // Для вызова межсервисных апи
    callGet: callGet,
    callPost: callPost,
    callPut: callPut,
    callDel: callDel,
    // Для вызова внешних апи
    callAxiGet: callAxiGet,
    callAxiPost: callAxiPost,
// Работа с событиями
    on: on,
    emit: emit,
    log: msLog,
    start: start,
};